/*
 * Simple single layer feed-forward neural network implementation
 * to self-train and subsequently recognise 8x8 bitmap characters
 *
 *  
 */

#include <stdio.h>

#define ACTIVATION_FUNCTION(x) (x)

extern unsigned char fontdata_8x8[];

int inputs[64];         /* the inputs (the bitmap) */
int weightings[8][64];  /* our input weightings */
int outputs[8];         /* our outputs */

/* simple font plotter */
void font_plot(int ch)
{
  unsigned char *c;   /* pointer for character data */
  int l=0;            /* line counter */

  c=&(fontdata_8x8[8*ch]); /* set pointer to first line of font */

  while(l!=8)
  {
    int bit=0; /* bit we are dealing with */

    while(bit!=8)
    {
      if(*c & (1 << bit))
        printf("#");
      else
        printf(" ");
      ++bit;
    }

    printf("\n");
    ++l;
    ++c;
  }
}

/* copies bitmap to neural network input */
void bitmap_to_input(int ch)
{
  unsigned char *c;   /* pointer for character data */
  int l=0;            /* line counter */
  int in=0;           /* index of the input we are setting */

  c=&(fontdata_8x8[8*ch]); /* set pointer to first line of font */

  while(l!=8)
  {
    int bit=0; /* bit we are dealing with */

    while(bit!=8)
    {
      if(*c & (1 << bit))
        inputs[in]=1;
      else
        inputs[in]=-1;
      ++bit;
      ++in;
    }

    ++l;
    ++c;
  }
}

/* prints a 8 bit binary number to screen */
void print_binary(int i)
{
  int bit=7;
 
  while(bit!=-1)
  {
    if(i & (1 << bit))
      printf("1");
    else
      printf("0");
    --bit;
  }  

  printf("\n");
}

/* converts an 8 bit integer to an array of 8 ints containing 0 and 1 */
void eightbits_to_ints(int ch, int *out)
{
  int bit=7;
 
  while(bit!=-1)
  {
    if(ch & (1 << bit))
      out[bit]=30;
    else
      out[bit]=0;
    --bit;
  }
}

/* converts an array of 8 ints containing 0 and 1 into an 8 bit int */
int ints_to_eightbits(int *input)
{
  int index=7, out=0;
 
  while(index!=-1)
  {
    if(ACTIVATION_FUNCTION(input[index]))
      out=out | (1 << (7-index));
    --index;
  }

  return out;
}

/* sets the weightings to -1 and +1 alternately */
void set_weightings(void)
{
  int j=0, k=0;

  while(j!=8)
  {
    k=0; 

    while(k!=64)
    {
      if(j+k % 2)
        weightings[j][k]=-1;
      else 
        weightings[j][k]=1;
      ++k;
    }

    ++j;
  }
}

/* neural_process */
int neural_process(void)
{
  int n=0; /* current neuron we are calculating */

  while(n!=8)
  {
    int level=0;
    int c=0;

    while(c!=64)
    {
      level=level+(weightings[n][c]*inputs[c]);
      ++c;
    }

    printf("%i,",level);

    //if(ACTIVATION_FUNCTION(level))
      outputs[n]=level;
      //else
      //outputs[n]=0;

    ++n;
  }

  printf("\n");

  return ints_to_eightbits(outputs);
}

/* attempt to learn - adjust weightings */
void neural_learn(int *doutcome)
{
  int n=0; /* current neuron we are calculating */

  while(n!=8)
  {
    int c=0;
    int e=0;

    e=doutcome[n]-outputs[n];

    while(c!=64)
    {
      if(e > 0)
        weightings[n][c]+=1;
      if(e < 0)
        weightings[n][c]-=1;
      printf("%i,",weightings[n][c]);
      ++c;
    }

    ++n;
  }

  printf("\n");  
}

/* attempt to learn a character */
void learn_character(int ch)
{
  int doutcome[8]; /* desired outcome */
  int nr;          /* returned from neural net */

  /* first copy the bitmap to the input array */
  bitmap_to_input(ch);

  /* copy the desired outcome (the ASCII of the character in binary) */
  eightbits_to_ints(ch,doutcome);

  /* now run through the neural net */
  nr=neural_process();

  printf("nr=%i\n");

  /* attempt to learn */
  neural_learn(doutcome);
}

int retrieve_character(int ch)
{
  int nr;          /* returned from neural net */

  /* first copy the bitmap to the input array */
  bitmap_to_input(ch);

  /* now run through the neural net */
  nr=neural_process();

  return nr;  
}

int main(void)
{
  int i=0, c=0;

  set_weightings();

  while(i!=1000)
  {
    c='A';
    while(c!='C')
    {
      learn_character(c);
      ++c;
    }
    ++i;
  }

  c='A';
  while(c!='C')
  {
    int x;

    x=retrieve_character(c);

    printf("input  "); print_binary(c);
    printf("%c (ASCII %i) retrieved from net as ASCII %i\n",c,c,x);
    printf("output "); print_binary(x);

    ++c;
  }

  return 0;
}
